{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Running on cuda\n"
     ]
    }
   ],
   "source": [
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "import torch\n",
    "import numpy as np\n",
    "import torch.optim as optim\n",
    "import pandas as pd\n",
    "from sklearn.model_selection import train_test_split\n",
    "\n",
    "device = \"cuda\" if torch.cuda.is_available() else \"cpu\"\n",
    "print(\"Running on\", device)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "495319"
      ]
     },
     "execution_count": 3,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "with open(\"gatunamn.txt\", \"r\", encoding=\"UTF-8\") as f:\n",
    "    names = f.read().split(\"\\n\")[2:]\n",
    "\n",
    "n_samples = len(names)\n",
    "n_samples"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "52"
      ]
     },
     "execution_count": 4,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "longest_name = max(names, key=len)\n",
    "longest_name = len(longest_name)\n",
    "longest_name"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "['D', 'H', '.', '3', 'è', 'F', '(', '7', 'Å', '-'] 93\n",
      "Tångavägen [32 25 70 54 52 38 24 54 48 70]\n"
     ]
    }
   ],
   "source": [
    "unique_chars = list(set(\"\".join(names + [\"#\"])))\n",
    "n_unique_chars = len(unique_chars)\n",
    "print(unique_chars[:10], n_unique_chars)\n",
    "\n",
    "c2i = {c: i for i, c in enumerate(unique_chars)}\n",
    "i2c = {i: c for i, c in enumerate(unique_chars)}\n",
    "\n",
    "number_dataset = [np.array([c2i[c] for c in name]) for name in names]\n",
    "print(names[0], number_dataset[0])\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "class AutoEncoderGatunamn(nn.Module):\n",
    "    def __init__(self, embed_size):\n",
    "        super(AutoEncoderGatunamn, self).__init__()\n",
    "\n",
    "        self._encoder = nn.Sequential(\n",
    "            nn.Linear(embed_size, 128),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(128, 64),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(64, 32),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(32, 16),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(16, 8),\n",
    "            nn.ReLU(),\n",
    "        )\n",
    "        self._decoder = nn.Sequential(\n",
    "            nn.Linear(8, 16),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(16, 32),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(32, 64),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(64, 128),\n",
    "            nn.ReLU(),\n",
    "            nn.Linear(128, embed_size),\n",
    "        )\n",
    "    \n",
    "    def forward(self, x):\n",
    "        x = self._encoder(x)\n",
    "        x = self._decoder(x)\n",
    "        return x\n",
    "    \n",
    "    def encode(self, x):\n",
    "        return self._encoder(x)\n",
    "    "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Tångavägen [32 25 70 54 52 38 24 54 48 70] torch.Size([62, 1])\n"
     ]
    }
   ],
   "source": [
    "pad_size = int(1.2 * longest_name)\n",
    "padded_number_dataset = [F.pad(torch.Tensor(name), (0, pad_size - len(name)), \"constant\", c2i[\"#\"]).view(-1, 1).long() for name in number_dataset]\n",
    "print(names[0], number_dataset[0], padded_number_dataset[0].shape)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "torch.Size([396255, 62]) torch.Size([99064, 62])\n"
     ]
    }
   ],
   "source": [
    "from torch.utils.data import DataLoader, Dataset\n",
    "\n",
    "\n",
    "class GatunamnDataset(Dataset):\n",
    "    def __init__(self, data):\n",
    "        self._data = data\n",
    "        self._pad_symbol = c2i[\"#\"]\n",
    "    \n",
    "    def __len__(self):\n",
    "        return len(self._data)\n",
    "    \n",
    "    def __getitem__(self, idx):\n",
    "        x = self._data[idx]\n",
    "        return x\n",
    "    \n",
    "\n",
    "X = torch.cat(padded_number_dataset, dim=1).float().T\n",
    "X = -1 + (X - X.min()) * 2 / (X.max() - X.min())\n",
    "X_train, X_test = train_test_split(X, test_size=0.2, random_state=42)\n",
    "print(X_train.shape, X_test.shape)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Epoch 1/10, Train loss: 0.019771, Test loss: 0.014532\n",
      "Epoch 2/10, Train loss: 0.013899, Test loss: 0.013421\n",
      "Epoch 3/10, Train loss: 0.012999, Test loss: 0.012390\n",
      "Epoch 4/10, Train loss: 0.011897, Test loss: 0.011301\n",
      "Epoch 5/10, Train loss: 0.011103, Test loss: 0.010972\n",
      "Epoch 6/10, Train loss: 0.010572, Test loss: 0.010272\n",
      "Epoch 7/10, Train loss: 0.010146, Test loss: 0.009963\n",
      "Epoch 8/10, Train loss: 0.009796, Test loss: 0.009562\n",
      "Epoch 9/10, Train loss: 0.009520, Test loss: 0.009207\n",
      "Epoch 10/10, Train loss: 0.009149, Test loss: 0.008572\n"
     ]
    }
   ],
   "source": [
    "model = AutoEncoderGatunamn(pad_size).to(device)\n",
    "optimizer = optim.Adam(model.parameters(), lr=0.001)\n",
    "criterion = nn.MSELoss()\n",
    "\n",
    "train_loader = DataLoader(GatunamnDataset(X_train), batch_size=128, shuffle=True)\n",
    "test_loader = DataLoader(GatunamnDataset(X_test), batch_size=32, shuffle=False)\n",
    "\n",
    "epochs = 10\n",
    "\n",
    "for epoch in range(epochs):\n",
    "    model.train()\n",
    "    running_train_loss = 0.0\n",
    "    running_test_loss = 0.0\n",
    "\n",
    "    for batch in train_loader:\n",
    "        optimizer.zero_grad()\n",
    "        batch = batch.to(device)\n",
    "        output = model(batch)\n",
    "        train_loss = criterion(output.view(-1), batch.float().view(-1))\n",
    "        running_train_loss += train_loss.item()\n",
    "        train_loss.backward()\n",
    "        optimizer.step()\n",
    "\n",
    "    model.eval()\n",
    "    with torch.no_grad():\n",
    "        for batch in test_loader:\n",
    "            batch = batch.to(device)\n",
    "            output = model(batch)\n",
    "            test_loss = criterion(output.view(-1), batch.float().view(-1))\n",
    "            running_test_loss += test_loss.item()\n",
    "\n",
    "    running_train_loss /= len(train_loader)\n",
    "    running_test_loss /= len(test_loader)\n",
    "\n",
    "    print(f\"Epoch {epoch + 1}/{epochs}, Train loss: {running_train_loss:.6f}, Test loss: {running_test_loss:.6f}\")\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "92 0\n"
     ]
    }
   ],
   "source": [
    "X_max = max(c2i.values())\n",
    "X_min = min(c2i.values())\n",
    "\n",
    "print(X_max, X_min)\n",
    "\n",
    "x = \"Duettväg\"\n",
    "xx = np.array([c2i[c] for c in x])\n",
    "xx = F.pad(torch.Tensor(xx), (0, pad_size - len(xx)), \"constant\", c2i[\"#\"]).view(-1, 1).long()\n",
    "x_test = -1 + (xx - X_min) * 2 / (X_max - X_min)\n",
    "x_test_out = model.encode(x_test.T.to(device)).cpu().detach().numpy()\n",
    "\n",
    "y = \"Duettvägen\"\n",
    "yy = np.array([c2i[c] for c in y])\n",
    "yy = F.pad(torch.Tensor(yy), (0, pad_size - len(yy)), \"constant\", c2i[\"#\"]).view(-1, 1).long()\n",
    "y_test = -1 + (yy - X_min) * 2 / (X_max - X_min)\n",
    "y_test_out = model.encode(y_test.T.to(device)).cpu().detach().numpy()\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "array([[-0.00142717,  0.00262201, -0.00808334,  0.        ,  0.        ,\n",
       "         0.        ,  0.        , -0.00378311]], dtype=float32)"
      ]
     },
     "execution_count": 11,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "x_test_out-y_test_out"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "myvenv",
   "language": "python",
   "name": "myvenv"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
